use aoc_runner_derive::{aoc, aoc_generator};
use nom::bytes::complete::tag;
use nom::character::complete::i64 as parse_u64;
use nom::multi::separated_list0;
use nom::IResult;

#[derive(Debug, Clone, Copy)]
struct Case {
    button_a: (i64, i64),
    button_b: (i64, i64),
    prize: (i64, i64),
}

type Input = Vec<Case>;

#[aoc_generator(day13)]
fn parse(input: &str) -> Input {
    separated_list0(tag("\n\n"), parse_case)(input).unwrap().1
}

fn parse_case(input: &str) -> IResult<&str, Case> {
    let (input, _) = tag("Button A: X+")(input)?;
    let (input, button_a_x) = parse_u64(input)?;
    let (input, _) = tag(", Y+")(input)?;
    let (input, button_a_y) = parse_u64(input)?;
    let (input, _) = tag("\nButton B: X+")(input)?;
    let (input, button_b_x) = parse_u64(input)?;
    let (input, _) = tag(", Y+")(input)?;
    let (input, button_b_y) = parse_u64(input)?;
    let (input, _) = tag("\nPrize: X=")(input)?;
    let (input, prize_x) = parse_u64(input)?;
    let (input, _) = tag(", Y=")(input)?;
    let (input, prize_y) = parse_u64(input)?;

    Ok((
        input,
        Case {
            button_a: (button_a_x, button_a_y),
            button_b: (button_b_x, button_b_y),
            prize: (prize_x, prize_y),
        },
    ))
}

#[aoc(day13, part1)]
fn part1(input: &Input) -> i64 {
    input
        .iter()
        .copied()
        .map(|case| {
            let vec_a = case.button_a;
            let vec_b = case.button_b;
            let vec_p = case.prize;

            if (vec_p.0 * vec_b.1 - vec_b.0 * vec_p.1) % (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1)
                == 0
                && (vec_a.0 * vec_p.1 - vec_p.0 * vec_a.1) % (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1)
                    == 0
            {
                let a = (vec_p.0 * vec_b.1 - vec_b.0 * vec_p.1)
                    / (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1);
                let b = (vec_a.0 * vec_p.1 - vec_p.0 * vec_a.1)
                    / (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1);
                a * 3 + b
            } else {
                0
            }
        })
        .sum()
}

#[aoc(day13, part2)]
fn part2(input: &Input) -> i64 {
    input
        .iter()
        .copied()
        .map(|case| {
            let vec_a = case.button_a;
            let vec_b = case.button_b;
            let vec_p = case.prize;
            let vec_p = (vec_p.0 + 10_000_000_000_000, vec_p.1 + 10_000_000_000_000);

            if (vec_p.0 * vec_b.1 - vec_b.0 * vec_p.1) % (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1)
                == 0
                && (vec_a.0 * vec_p.1 - vec_p.0 * vec_a.1) % (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1)
                    == 0
            {
                let a = (vec_p.0 * vec_b.1 - vec_b.0 * vec_p.1)
                    / (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1);
                let b = (vec_a.0 * vec_p.1 - vec_p.0 * vec_a.1)
                    / (vec_a.0 * vec_b.1 - vec_b.0 * vec_a.1);
                a * 3 + b
            } else {
                0
            }
        })
        .sum()
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(
                "Button A: X+94, Y+34
Button B: X+22, Y+67
Prize: X=8400, Y=5400"
            )),
            280
        );
        assert_eq!(
            part1(&parse(
                "Button A: X+94, Y+34
Button B: X+22, Y+67
Prize: X=8400, Y=5400

Button A: X+26, Y+66
Button B: X+67, Y+21
Prize: X=12748, Y=12176

Button A: X+17, Y+86
Button B: X+84, Y+37
Prize: X=7870, Y=6450

Button A: X+69, Y+23
Button B: X+27, Y+71
Prize: X=18641, Y=10279"
            )),
            480
        );
        assert_eq!(
            part1(&parse(include_str!("../input/2024/day13.txt").trim_end())),
            28753
        );
    }

    #[test]
    fn part2_example() {
        assert_eq!(
            part2(&parse(include_str!("../input/2024/day13.txt").trim_end())),
            102718967795500
        );
    }
}
