use aoc_runner_derive::{aoc, aoc_generator};

use nom::{
    bytes::complete::tag,
    character::complete::{i64 as parse_i64, newline, space1},
    error::Error,
    multi::separated_list0,
    sequence::{preceded, separated_pair},
};

use num_modular::ModularPow;

type Input = Vec<((i64, i64), (i64, i64))>;

#[aoc_generator(day14)]
fn parse(input: &str) -> Input {
    separated_list0(
        newline::<&str, Error<&str>>,
        separated_pair(
            preceded(tag("p="), separated_pair(parse_i64, tag(","), parse_i64)),
            space1,
            preceded(tag("v="), separated_pair(parse_i64, tag(","), parse_i64)),
        ),
    )(input)
    .unwrap()
    .1
}

#[aoc(day14, part1)]
fn part1(input: &Input) -> i64 {
    #[cfg(test)]
    let (w, h) = (7, 11);
    #[cfg(not(test))]
    let (w, h) = (101, 103);

    let mut quadrants = [0, 0, 0, 0];

    input.iter().copied().for_each(|(p, v)| {
        let (px, py) = (p.0, p.1);
        let (vx, vy) = (v.0 + w, v.1 + h);
        let (ex, ey) = ((px + vx * 100) % w, (py + vy * 100) % h);
        if ex < w / 2 && ey < h / 2 {
            quadrants[0] += 1;
        } else if ex > (w - 1) / 2 && ey < h / 2 {
            quadrants[1] += 1;
        } else if ex < w / 2 && ey > (h - 1) / 2 {
            quadrants[2] += 1;
        } else if ex > (w - 1) / 2 && ey > (h - 1) / 2 {
            quadrants[3] += 1;
        }
    });

    quadrants.iter().product()
}

#[aoc(day14, part2)]
fn part2(input: &Input) -> i64 {
    let (w, h) = (101, 103);

    let mut offset_x = 0;
    let mut min_var = f64::INFINITY;
    for t in 0..w {
        let values = input.iter().map(|((p, _), (v, _))| (p + t * (v + w)) % w);
        let mean = values.clone().sum::<i64>() as f64 / input.len() as f64;
        let variance = values
            .map(|v| (v as f64 - mean) * (v as f64 - mean))
            .sum::<f64>()
            / input.len() as f64;
        if variance < min_var {
            min_var = variance;
            offset_x = t;
        }
    }

    let mut offset_y = 0;
    let mut min_var = f64::INFINITY;
    for t in 0..h {
        let values = input.iter().map(|((_, p), (_, v))| (p + t * (v + h)) % h);
        let mean = values.clone().sum::<i64>() as f64 / input.len() as f64;
        let variance = values
            .map(|v| (v as f64 - mean) * (v as f64 - mean))
            .sum::<f64>()
            / input.len() as f64;
        if variance < min_var {
            min_var = variance;
            offset_y = t;
        }
    }

    let invmod_x = (w as u64).powm((h - 2) as u64, &(h as u64)) as i64;
    let invmod_y = (h as u64).powm((w - 2) as u64, &(w as u64)) as i64;
    (offset_x * h * invmod_x + offset_y * w * invmod_y) % (w * h)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(
                "p=0,4 v=3,-3
p=6,3 v=-1,-3
p=10,3 v=-1,2
p=2,0 v=2,-1
p=0,0 v=1,3
p=3,0 v=-2,-2
p=7,6 v=-1,-3
p=3,0 v=-1,-2
p=9,3 v=2,3
p=7,3 v=-1,2
p=2,4 v=2,-3
p=9,5 v=-3,-3"
            )),
            12
        );
    }

    #[test]
    fn part2_example() {
        assert_eq!(
            part2(&parse(include_str!("../input/2024/day14.txt").trim_end())),
            6398
        );
    }
}
