use aoc_runner_derive::{aoc, aoc_generator};
use nom::bytes::complete::tag;
use nom::character::complete::i64 as parse_i64;
use nom::error::Error;
use nom::multi::separated_list1;
use nom::sequence::{preceded, separated_pair, tuple};

type Input = ((i64, i64, i64), Vec<i64>);

#[aoc_generator(day17)]
fn parse(input: &str) -> Input {
    separated_pair(
        tuple((
            preceded(tag("Register A: "), parse_i64::<&str, Error<&str>>),
            preceded(tag("\nRegister B: "), parse_i64),
            preceded(tag("\nRegister C: "), parse_i64),
        )),
        tag("\n\n"),
        preceded(tag("Program: "), separated_list1(tag(","), parse_i64)),
    )(input)
    .unwrap()
    .1
}

#[aoc(day17, part1)]
fn part1(input: &Input) -> String {
    let mut registers = input.0;
    let program = input.1.clone();

    let mut pc = 0;

    let mut output = Vec::new();

    loop {
        if pc >= program.len() {
            break;
        }

        let instruction = program[pc];
        let literal_value = program[pc + 1];
        let combo_value = match program[pc + 1] {
            0 => 0,
            1 => 1,
            2 => 2,
            3 => 3,
            4 => registers.0,
            5 => registers.1,
            6 => registers.2,
            _ => unreachable!(),
        };

        match instruction {
            0 => registers.0 >>= combo_value,
            1 => registers.1 ^= literal_value,
            2 => registers.1 = combo_value % 8,
            3 => {
                if registers.0 != 0 {
                    pc = literal_value as usize;
                    continue;
                }
            }
            4 => registers.1 ^= registers.2,
            5 => output.push(combo_value % 8),
            6 => registers.1 = registers.0 >> combo_value,
            7 => registers.2 = registers.0 >> combo_value,
            _ => unreachable!(),
        }

        pc += 2;
    }

    output
        .iter()
        .map(|x| x.to_string())
        .collect::<Vec<String>>()
        .join(",")
}

#[aoc(day17, part2)]
fn part2(input: &Input) -> i64 {
    let program = input.1.clone();
    let mut candidates = vec![0];

    for target in program.iter().rev() {
        let mut new_candidates = vec![];
        for candidate in candidates {
            let a = candidate;
            new_candidates.extend(
                (0..8)
                    .filter(|b| {
                        let a = (a << 3) + b;
                        let b = a % 8;
                        let b = b ^ 5;
                        let c = a >> b;
                        let b = b ^ c;
                        let b = b ^ 6;

                        b % 8 == *target
                    })
                    .map(|b| (a << 3) + b),
            );
        }
        candidates = new_candidates;
    }

    candidates[0]
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(
                "Register A: 729
Register B: 0
Register C: 0

Program: 0,1,5,4,3,0"
            )),
            "4,6,3,5,6,3,5,2,1,0"
        );
    }
}
