use std::collections::VecDeque;

use aoc_runner_derive::{aoc, aoc_generator};
use nom::{
    bytes::complete::tag,
    character::complete::{i64 as parse_i64, newline},
    error::Error,
    multi::separated_list0,
    sequence::separated_pair,
};

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum Cell {
    Empty,
    Byte,
    Visited,
}

type Input = Vec<(i64, i64)>;

#[aoc_generator(day18)]
fn parse(input: &str) -> Input {
    separated_list0(
        newline::<&str, Error<&str>>,
        separated_pair(parse_i64, tag(","), parse_i64),
    )(input)
    .unwrap()
    .1
}

#[aoc(day18, part1)]
fn part1(input: &Input) -> usize {
    #[cfg(not(test))]
    let (w, h, bytes) = (71, 71, &input[..1024]);
    #[cfg(test)]
    let (w, h, bytes) = (7, 7, &input[..12]);

    let mut map = vec![vec![Cell::Empty; w]; h];

    bytes
        .iter()
        .for_each(|(x, y)| map[*y as usize][*x as usize] = Cell::Byte);

    let mut deque = VecDeque::new();
    deque.push_back((0, 0, 0));

    while let Some((x, y, depth)) = deque.pop_front() {
        if x == w - 1 && y == h - 1 {
            return depth;
        }

        if map[y][x] == Cell::Visited {
            continue;
        }

        map[y][x] = Cell::Visited;

        if x >= 1 && map[y][x - 1] == Cell::Empty {
            deque.push_back((x - 1, y, depth + 1));
        }
        if x < w - 1 && map[y][x + 1] == Cell::Empty {
            deque.push_back((x + 1, y, depth + 1));
        }
        if y >= 1 && map[y - 1][x] == Cell::Empty {
            deque.push_back((x, y - 1, depth + 1));
        }
        if y < h - 1 && map[y + 1][x] == Cell::Empty {
            deque.push_back((x, y + 1, depth + 1));
        }
    }

    unreachable!()
}

#[aoc(day18, part2)]
fn part2(input: &Input) -> String {
    #[cfg(not(test))]
    let (w, h, min_idx) = (71, 71, 1024);
    #[cfg(test)]
    let (w, h, min_idx) = (7, 7, 12);

    let (mut l, mut r) = (min_idx - 1, input.len() - 1);

    while r - l > 1 {
        let mut possible = false;

        let mut map = vec![vec![Cell::Empty; w]; h];

        let bytes = (l + r) >> 1;

        input[0..bytes]
            .iter()
            .for_each(|(x, y)| map[*y as usize][*x as usize] = Cell::Byte);

        let mut deque = VecDeque::new();
        deque.push_back((0, 0, 0));

        while let Some((x, y, depth)) = deque.pop_front() {
            if x == w - 1 && y == h - 1 {
                possible = true;
                break;
            }

            if map[y][x] == Cell::Visited {
                continue;
            }

            map[y][x] = Cell::Visited;

            if x >= 1 && map[y][x - 1] == Cell::Empty {
                deque.push_back((x - 1, y, depth + 1));
            }
            if x < w - 1 && map[y][x + 1] == Cell::Empty {
                deque.push_back((x + 1, y, depth + 1));
            }
            if y >= 1 && map[y - 1][x] == Cell::Empty {
                deque.push_back((x, y - 1, depth + 1));
            }
            if y < h - 1 && map[y + 1][x] == Cell::Empty {
                deque.push_back((x, y + 1, depth + 1));
            }
        }

        if possible {
            l = bytes;
        } else {
            r = bytes;
        }
    }

    format!("{},{}", input[l].0, input[l].1).to_string()
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(
            part1(&parse(
                "5,4
4,2
4,5
3,0
2,1
6,3
2,4
1,5
0,6
3,3
2,6
5,1
1,2
5,5
2,5
6,5
1,4
0,4
6,4
1,1
6,1
1,0
0,5
1,6
2,0"
            )),
            22
        );
    }

    #[test]
    fn part2_example() {
        assert_eq!(
            part2(&parse(
                "5,4
4,2
4,5
3,0
2,1
6,3
2,4
1,5
0,6
3,3
2,6
5,1
1,2
5,5
2,5
6,5
1,4
0,4
6,4
1,1
6,1
1,0
0,5
1,6
2,0"
            )),
            "6,1"
        );
    }
}
