use std::collections::HashMap;

use aoc_runner_derive::{aoc, aoc_generator};
use nom::character::complete::u64 as parse_u64;
use nom::multi::separated_list1;
use nom::IResult;
use nom::{
    bytes::complete::tag, character::complete::newline, multi::separated_list0,
    sequence::separated_pair,
};
use topological_sort::TopologicalSort;

type Input = (Vec<(u64, u64)>, Vec<Vec<u64>>);

#[aoc_generator(day5)]
fn parse(input: &str) -> Input {
    parse_lines(input).unwrap().1
}

fn parse_lines(input: &str) -> IResult<&str, Input> {
    let (input, rules) =
        separated_list0(newline, separated_pair(parse_u64, tag("|"), parse_u64))(input)?;
    let (input, _) = newline(input)?;
    let (input, _) = newline(input)?;
    let (input, records) = separated_list0(newline, separated_list1(tag(","), parse_u64))(input)?;
    Ok((input, (rules, records)))
}

#[aoc(day5, part1)]
fn part1(input: &Input) -> u64 {
    let rules = input.0.clone();
    let records = input.1.clone();

    let mut rules_map: HashMap<u64, Vec<u64>> = HashMap::new();
    rules.iter().for_each(|rule| {
        if let std::collections::hash_map::Entry::Vacant(e) = rules_map.entry(rule.1) {
            e.insert(vec![rule.0]);
        } else {
            rules_map.get_mut(&rule.1).unwrap().push(rule.0);
            rules_map.get_mut(&rule.1).unwrap().sort();
        }
    });

    let mut correct = 0;
    records.iter().for_each(|record| {
        let mut forbidden = vec![];
        let valid = record.iter().all(|page| {
            let allowed = !forbidden.contains(page);
            if rules_map.contains_key(page) {
                let mut new_forbidden = rules_map.get(page).unwrap().clone();
                forbidden.append(&mut new_forbidden);
            }
            allowed
        });
        if valid {
            correct += record[record.len() >> 1]
        }
    });

    correct
}

#[aoc(day5, part2)]
fn part2(input: &Input) -> u64 {
    let rules = input.0.clone();
    let records = input.1.clone();

    let mut rules_map: HashMap<u64, Vec<u64>> = HashMap::new();
    rules.iter().for_each(|rule| {
        if let std::collections::hash_map::Entry::Vacant(e) = rules_map.entry(rule.1) {
            e.insert(vec![rule.0]);
        } else {
            rules_map.get_mut(&rule.1).unwrap().push(rule.0);
            rules_map.get_mut(&rule.1).unwrap().sort();
        }
    });

    let mut correct = 0;
    records.iter().for_each(|record| {
        let mut forbidden = vec![];
        let valid = record.iter().all(|page| {
            let allowed = !forbidden.contains(page);
            if rules_map.contains_key(page) {
                let mut new_forbidden = rules_map.get(page).unwrap().clone();
                forbidden.append(&mut new_forbidden);
            }
            allowed
        });
        if !valid {
            // dbg!(&record);
            let mut ts = TopologicalSort::<u64>::new();
            for i in record {
                ts.insert(*i);
                if let Some(rules) = rules_map.get(i) {
                    for rule in rules {
                        ts.add_dependency(*rule, *i);
                    }
                }
            }

            let sorted = ts
                .clone()
                .filter(|a| record.contains(a))
                .collect::<Vec<u64>>();
            correct += sorted[sorted.len() >> 1]
        }
    });

    correct
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn part1_example() {
        assert_eq!(part1(&parse("47|53\n97|13\n97|61\n97|47\n75|29\n61|13\n75|53\n29|13\n97|29\n53|29\n61|53\n97|53\n61|29\n47|13\n75|47\n97|75\n47|61\n75|61\n47|29\n75|13\n53|13\n\n75,47,61,53,29\n97,61,53,29,13\n75,29,13\n75,97,47,61,53\n61,13,29\n97,13,75,29,47")), 143);
    }

    #[test]
    fn part2_example() {
        assert_eq!(part2(&parse("47|53\n97|13\n97|61\n97|47\n75|29\n61|13\n75|53\n29|13\n97|29\n53|29\n61|53\n97|53\n61|29\n47|13\n75|47\n97|75\n47|61\n75|61\n47|29\n75|13\n53|13\n\n75,47,61,53,29\n97,61,53,29,13\n75,29,13\n75,97,47,61,53\n61,13,29\n97,13,75,29,47")), 123);
    }
}
